# Copyright 2021, 2022 Milan Meulemans.
#
# This file is part of aioaseko.
#
# aioaseko is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# aioaseko is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with aioaseko.  If not, see <https://www.gnu.org/licenses/>.

"""aioAseko web API account."""
from __future__ import annotations

from dataclasses import dataclass

from aiohttp import ClientError, ClientResponse, ClientSession

from .exceptions import APIUnavailable, InvalidAuthCredentials
from .unit import Unit


class WebAccount:
    """Aseko web account."""

    def __init__(self, session: ClientSession, username: str, password: str) -> None:
        """Init Aseko account."""
        self._session = session
        self._username = username
        self._password = password

    async def _request(
        self, method: str, path: str, data: dict | None = None
    ) -> ClientResponse:
        """Make a request to the Aseko web API."""
        resp = await self._session.request(
            method, f"https://pool.aseko.com/api/{path}", data=data
        )
        if resp.status == 401:
            raise InvalidAuthCredentials
        try:
            resp.raise_for_status()
        except ClientError:
            raise APIUnavailable
        return resp

    async def login(self) -> AccountInfo:
        """Login to the Aseko web API."""
        resp = await self._request(
            "post",
            "login",
            {
                "username": self._username,
                "password": self._password,
                "agree": "on",
            },
        )
        data = await resp.json()
        return AccountInfo(data["email"], data["userId"], data.get("language"))

    async def get_units(self) -> list[Unit]:
        """Get units."""
        resp = await self._request("get", "units")
        data = await resp.json()
        return [
            Unit(
                self,
                int(item["serialNumber"]),
                item["type"],
                item.get("name"),
                item.get("notes"),
                item["timezone"],
                item["isOnline"],
                item["dateLastData"],
                item["hasError"],
            )
            for item in data["items"]
        ]


@dataclass(frozen=True)
class AccountInfo:
    """Aseko account info."""

    email: str
    user_id: str
    language: str | None



# Copyright 2021, 2022 Milan Meulemans.
#
# This file is part of aioaseko.
#
# aioaseko is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# aioaseko is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with aioaseko.  If not, see <https://www.gnu.org/licenses/>.

"""aioAseko mobile API account."""
from __future__ import annotations

from typing import TYPE_CHECKING

from aiohttp import ClientError

from .exceptions import APIUnavailable, InvalidAuthCredentials
from .unit import Unit

from jwt import decode, get_unverified_header
from time import time


if TYPE_CHECKING:
    from aiohttp import ClientResponse, ClientSession

TOKEN_EXPIRATION_OFFSET = 60

class MobileAccount:
    """Aseko account."""
    
    _access_token: str | None = None
    _access_token_expiration: int | None = None

    def __init__(
        self,
        session: ClientSession,
        username: str | None = None,
        password: str | None = None,
        refresh_token: str | None = None,
    ) -> None:
        """Init Aseko account."""
        self._session = session
        self._username = username
        self._password = password
        self._refresh_token = refresh_token

    @property
    def refresh_token(self) -> str | None:
        """Return refresh token."""
        return self._refresh_token

    async def _request(
        self, method: str, path: str, data: dict | None = None, auth: bool = True
    ) -> ClientResponse:
        """Make a request to the Aseko mobile API."""
        resp = await self._session.request(
            method,
            f"https://pool.aseko.com/api/v2/{path}",
            data=data,
            headers=None if not auth
            else {"access-token": await self._get_valid_access_token()},
        )
        if resp.status == 401:
            raise InvalidAuthCredentials
        try:
            resp.raise_for_status()
        except ClientError:
            raise APIUnavailable
        return resp

    async def login(self) -> None:
        """Login to Aseko Pool Live with username and password."""
        resp = await self._request(
            "post",
            "login",
            {
                "username": self._username,
                "password": self._password,
                "firebaseId": "",
            },
            False,
        )
        data = await resp.json()
        self._retrieve_tokens(data)

    async def _get_valid_access_token(self) -> str:
        """Return a valid access token."""
        if self._access_token is not None:
            assert self._access_token_expiration is not None
            if self._access_token_expiration >= time() + TOKEN_EXPIRATION_OFFSET:
                return self._access_token
            else:
                self._access_token = None
        if self._refresh_token is not None:
            try:
                await self._refresh()
            except InvalidAuthCredentials:
                self._refresh_token = None
            else:
                assert self._access_token is not None
                return self._access_token
        if self._username is not None and self._password is not None:
            await self.login()
            assert self._access_token is not None
            return self._access_token
        raise InvalidAuthCredentials("No valid refresh token or username and password available.")

    async def _refresh(self) -> None:
        """Refresh access token for Aseko Pool Live with refresh token."""
        resp = await self._request(
            "post",
            "refresh",
            {
                "refreshToken": self._refresh_token,
                "firebaseId": "",
            },
            False,
        )
        data = await resp.json()
        self._retrieve_tokens(data)

    async def logout(self) -> None:
        """Logout Aseko Pool Live account."""
        await self._request("post", "logout")
        self._access_token = None
        self._access_token_expiration = None
        self._refresh_token = None

    async def get_units(self) -> list[Unit]:
        """Get units."""
        resp = await self._request("get", "units")
        data = await resp.json()
        return [
            Unit(
                self,
                int(item["serialNumber"]),
                item["type"],
                item.get("name"),
                item.get("notes"),
                item["timezone"],
                item["isOnline"],
                item["dateLastData"],
                item["hasError"],
            )
            for item in data["items"]
        ]

    def _retrieve_tokens(self, data: dict) -> None:
        algorithm = get_unverified_header(data["accessToken"]).get('alg')
        token = decode(
            jwt=data["accessToken"],
            key="",
            algorithms=algorithm,
            options={
                "verify_signature": False
            },
        )
        self._access_token = data["accessToken"]
        self._access_token_expiration = token["exp"]
        self._refresh_token = data["refreshToken"]



# Copyright 2021, 2022 Milan Meulemans.
#
# This file is part of aioaseko.
#
# aioaseko is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# aioaseko is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with aioaseko.  If not, see <https://www.gnu.org/licenses/>.

"""aioAseko exceptions."""


class InvalidAuthCredentials(Exception):
    """Invalid authentication credentials."""


class APIUnavailable(Exception):
    """API Unavailable."""



# Copyright 2021, 2022, 2024 Milan Meulemans.
#
# This file is part of aioaseko.
#
# aioaseko is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# aioaseko is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with aioaseko.  If not, see <https://www.gnu.org/licenses/>.

"""aioAseko variable."""
from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class Variable:
    """Aseko variable."""

    type: str
    name: str
    unit: str
    icon: str | None
    color: str | None
    has_error: bool
    current_value: int | None
    required_value: int | None
    has_alarm: bool | None
    min_value: int | None
    max_value: int | None



# Copyright 2021, 2022, 2024 Milan Meulemans.
#
# This file is part of aioaseko.
#
# aioaseko is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# aioaseko is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with aioaseko.  If not, see <https://www.gnu.org/licenses/>.

"""aioAseko unit."""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from .variable import Variable

if TYPE_CHECKING:
    from .mobile import MobileAccount
    from .web import WebAccount


class Unit:
    """Aseko unit."""

    def __init__(
        self,
        account: MobileAccount | WebAccount,
        serial_number: int,
        type: str,
        name: str | None,
        notes: str | None,
        timezone: str,
        is_online: bool,
        date_last_data: str,
        has_error: bool,
    ) -> None:
        """Init Aseko unit."""
        self._account = account
        self._serial_number = serial_number
        self._type = type
        self._name = name
        self._notes = notes
        self._timezone = timezone
        self._is_online = is_online
        self._date_last_data = date_last_data
        self._has_error = has_error

    @property
    def serial_number(self) -> int:
        """Return serial number, set on init."""
        return self._serial_number

    @property
    def type(self) -> str:
        """Return type, set on init."""
        return self._type

    @property
    def name(self) -> str | None:
        """Return name, set on init."""
        return self._name

    @property
    def notes(self) -> str | None:
        """Return notes, set on init."""
        return self._notes

    @property
    def timezone(self) -> str:
        """Return timezone, set on init."""
        return self._timezone

    @property
    def is_online(self) -> bool:
        """Return is online, set on init."""
        return self._is_online

    @property
    def date_last_data(self) -> str:
        """Return date of last data, set on init."""
        return self._date_last_data

    @property
    def has_error(self) -> bool:
        """Return has error, set on init and get_state()."""
        return self._has_error

    @property
    def errors(self) -> list[UnitError]:
        """Return errors, set on get_state()."""
        return self._errors

    @property
    def variables(self) -> list[Variable]:
        """Return variables, set on get_state()."""
        return self._variables

    @property
    def has_alarm(self) -> bool:
        """Return has alarm, set on get_state()."""
        return self._has_alarm

    @property
    def water_flow(self) -> bool:
        """Return water flow, set on get_state()."""
        return self._water_flow

    async def get_state(self) -> None:
        """Get the unit state."""
        resp = await self._account._request("get", f"units/{self._serial_number}")
        data = await resp.json()
        self._errors = [
            UnitError(error["type"], error["title"], error["content"])
            for error in data["errors"]
        ]
        self._has_error = len(self._errors) >= 1
        self._variables = [
            Variable(
                variable["type"],
                variable["name"],
                variable["unit"],
                variable.get("icon"),
                variable.get("color"),
                variable["hasError"],
                variable.get("currentValue"),
                variable.get("required"),
                variable.get("alarm", {}).get("active"),
                variable.get("alarm", {}).get("minValue"),
                variable.get("alarm", {}).get("maxValue"),
            )
            for variable in data["variables"]
        ]
        self._has_alarm: bool = data["errorsAlarm"]["active"]
        self._water_flow: bool = not data.get("noWaterFlow", False)


@dataclass(frozen=True)
class UnitError:
    """Aseko unit error."""

    type: str
    title: str
    content: list[str]



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="aioaseko",
    version="0.2.0",
    author="Milan Meulemans",
    author_email="milan.meulemans@live.be",
    description="Async Python package for the Aseko Pool Live API",
    keywords="aseko pool live api asin aqua",
    license="LGPLv3+",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/milanmeu/aioaseko",
    project_urls={
        "Say Thanks!": "https://saythanks.io/to/milan.meulemans@live.be",
        "Bug Tracker": "https://github.com/milanmeu/aioaseko/issues",
        "Source Code": "https://github.com/milanmeu/aioaseko",
        "Documentation": "https://github.com/milanmeu/aioaseko/blob/main/README.md",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Topic :: Home Automation",
        "Typing :: Typed",
    ],
    python_requires=">=3.8",
    packages=["aioaseko"],
    package_data={"aioaseko": ["py.typed"]},
    install_requires=["aiohttp", "pyjwt"]
)
